// src/manip.rs

use std::{collections::HashMap, fs, path::PathBuf};

use log::error;

// ENUMS

/// Specifies a limit type, whether block or silence
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum Limit {
    Block,
    Silence,
}

// STRUCTS

/// Wraps a host's weighted limits
#[derive(Debug, Default, Eq, PartialEq)]
pub struct LimitIndices {
    pub block: u16,
    pub silence: u16,
}

impl LimitIndices {
    pub fn add_limit(&mut self, limit: Limit, weight: u16) -> &mut Self {
        match limit {
            Limit::Block => self.block += weight,
            Limit::Silence => self.silence += weight,
        }
        self
    }

    pub fn normalize(&self, max: (u16, u16)) -> (u16, u16) {
        let block = (self.block as f32) / (max.0 as f32) * 100f32;
        let mute = (self.silence as f32) / (max.1 as f32) * 100f32;

        (block as u16, mute as u16)
    }
}

impl From<(u16, u16)> for LimitIndices {
    /// Creates indices from a tuple of two `u16` weights. Useful mostly for
    /// testing.
    fn from(value: (u16, u16)) -> Self {
        Self {
            block: value.0,
            silence: value.1,
        }
    }
}

/// Contains a mapping of hosts to limits as well as a trust value which is
/// used to weight limits when building a merged list
#[derive(Debug, PartialEq, Eq)]
pub struct LimitList {
    pub hosts: Vec<String>,
    pub trust: u16,
}

impl LimitList {
    pub fn from_file(path: PathBuf) -> std::io::Result<Self> {
        let mut list = LimitList {
            hosts: vec![],
            trust: 100,
        };
        let contents = fs::read_to_string(path)?;

        for host in contents.lines().filter(|line| !line.is_empty()) {
            list.hosts.push(host.to_string());
        }

        Ok(list)
    }
}

/// A map of hosts (as strings) to their limit weights
#[derive(Debug, Default, Eq, PartialEq)]
pub struct MergedLimitList {
    pub hostmap: HashMap<String, LimitIndices>,
    pub trusts: (u16, u16),
}

impl MergedLimitList {
    pub fn new() -> Self {
        Self {
            hostmap: HashMap::new(),
            trusts: (0, 0),
        }
    }

    pub fn add_limit_list(&mut self, src: LimitList, limit_type: Limit) -> &mut Self {
        for host in src.hosts.into_iter() {
            let entry = self.hostmap.entry(host).or_default();
            entry.add_limit(limit_type, src.trust);
        }

        match limit_type {
            Limit::Block => self.trusts.0 += src.trust,
            Limit::Silence => self.trusts.1 += src.trust,
        }

        self
    }

    pub fn export_file(
        self,
        block_path: &str,
        mute_path: &str,
        indices: (u16, u16),
    ) -> std::io::Result<()> {
        if self.hostmap.is_empty() {
            error!("Nothing to export!");
            return Ok(());
        }

        let (block_thresh, mute_thresh) = indices;
        let mut block_output: String = String::default();
        let mut mute_output: String = String::default();

        for item in self.hostmap.into_iter() {
            let (block_trust, mute_trust) = item.1.normalize(self.trusts);

            // TODO What should the behavior be when an instance is blocked or
            // muted by most instances, but neither rises above the threshold?
            if block_trust >= block_thresh {
                block_output.push_str(&(item.0.clone() + "\n"));
            }

            if mute_trust >= mute_thresh {
                mute_output.push_str(&(item.0.clone() + "\n"));
            }
        }

        fs::write(block_path, alphabetize_lines(block_output))?;
        fs::write(mute_path, alphabetize_lines(mute_output))?;

        Ok(())
    }
}

pub fn alphabetize_lines(string: String) -> String {
    let mut v: Vec<&str> = string.lines().collect();
    v.sort_by_key(|a| a.to_lowercase());

    v.join("\n") + "\n"
}
